#include "handmade.h"
#include "handmade_intrinsics.h"

internal void drawRectangle(GameOffscreenBuffer *buffer, real32 realMinX, real32 realMinY, real32 realMaxX, real32 realMaxY, real32 R, real32 G, real32 B) {
    int32 minX = roundReal32ToInt32(realMinX);
    int32 minY = roundReal32ToInt32(realMinY);
    int32 maxX = roundReal32ToInt32(realMaxX);
    int32 maxY = roundReal32ToInt32(realMaxY);

    uint32 color = (roundReal32ToUInt32(R*255.0f) << 16) |
                   (roundReal32ToUInt32(G*255.0f) <<  8) |
                   (roundReal32ToUInt32(B*255.0f) <<  0);

    if (minX < 0) {
        minX = 0;
    }

    if (minY < 0) {
        minY = 0;
    }

    if (maxX > buffer->width) {
        maxX = buffer->width;
    }

    if (maxY > buffer->height) {
        maxY = buffer->height;
    }

    uint8 *endOfBuffer = (uint8 *)buffer->memory + buffer->pitch*buffer->height;
    uint8 *row = ((uint8 *)buffer->memory + minX*buffer->bytesPerPixel + minY*buffer->pitch);
    for (int y = minY; y < maxY; y++) {
        uint32 *pixel = (uint32 *)row;
        for (int x = minX; x < maxX; x++) {
            *pixel++ = color;
        }
        row += buffer->pitch;
    }
}

internal void outputSound(GameSoundOutputBuffer *soundBuffer, GameState *state) {
    int16 toneVolume = 3000;
    int16 *sampleOut = soundBuffer->samples;
    for (int sampleIndex = 0; sampleIndex < soundBuffer->sampleCount; sampleIndex++) {
        int16 sampleValue = 0;
        *sampleOut++ = sampleValue;
        *sampleOut++ = sampleValue;
    }
}

inline uint32 getTileValueUnchecked(World *world, TileMap *tileMap, int32 testTileX, int32 testTileY) {
    Assert(tileMap);
    Assert(
        (testTileX >= 0) && (testTileX < world->tileCountX) &&
        (testTileY >= 0) && (testTileY < world->tileCountY)
    );
    return tileMap->tiles[testTileY * world->tileCountX + testTileX];
}

inline TileMap *getTileMap(World *world, int32 tileMapX, int32 tileMapY) {
    TileMap *tileMap = NULL;

    if ((tileMapX >= 0) && (tileMapX < world->tileMapCountX) &&
        (tileMapY >= 0) && (tileMapY < world->tileMapCountY)
    ) {
        tileMap = &world->tileMaps[tileMapY * world->tileMapCountX + tileMapX];
    }

    return tileMap;
}

inline void recanonicaliseOrd(World *world, int32 tileCount, int32 *tileMapOrd, int32 *tileOrd, real32* ord) {
    int32 offset = floorReal32ToInt32(*ord / world->tileSideInPixels);
    *tileOrd += offset;
    *ord -= offset*world->tileSideInPixels;

    Assert(*ord >= 0);
    Assert(*ord < world->tileSideInPixels);

    if (*tileOrd < 0) {
        *tileOrd += tileCount;
        *tileMapOrd -= 1;
    }

    if (*tileOrd >= tileCount) {
        *tileOrd -= tileCount;
        *tileMapOrd += 1;
    }
}

inline CanonicalPosition recanonicalisePosition(World* world, CanonicalPosition pos) {
    CanonicalPosition result = pos;

    recanonicaliseOrd(world, world->tileCountX, &result.tileMapX, &result.tileX, &result.x);
    recanonicaliseOrd(world, world->tileCountY, &result.tileMapY, &result.tileY, &result.y);

    return result;
}

internal bool32 isWorldPointEmpty(World *world, CanonicalPosition testPos) {
    bool32 isEmpty = false;

    TileMap *tileMap = getTileMap(world, testPos.tileMapX, testPos.tileMapY);
    isEmpty = (getTileValueUnchecked(world, tileMap, testPos.tileX, testPos.tileY) == 0);

    return isEmpty;
}

extern "C" GAME_UPDATE_AND_RENDER(gameUpdateAndRender) {
    Assert(sizeof(GameState) <= memory->permanentStorageSize);

    GameState *state = (GameState*)memory->permanentStorage;

    if (!memory->isInitialised) {
        state->playerPos.x = 150.0f;
        state->playerPos.y = 150.0f;
        state->playerPos.tileMapX = 0;
        state->playerPos.tileMapY = 0;

        memory->isInitialised = true;
    }

    const int TILEMAP_WIDTH = 16;
    const int TILEMAP_HEIGHT = 9;

    const int WORLD_WIDTH = 2;
    const int WORLD_HEIGHT = 2;

    uint32 tiles00[TILEMAP_HEIGHT][TILEMAP_WIDTH] = {
        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 1, 1, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
    };

    uint32 tiles01[TILEMAP_HEIGHT][TILEMAP_WIDTH] = {
        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
    };

    uint32 tiles10[TILEMAP_HEIGHT][TILEMAP_WIDTH] = {
        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
        {1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
        {1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
    };

    uint32 tiles11[TILEMAP_HEIGHT][TILEMAP_WIDTH] = {
        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
        {1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
        {0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
        {1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 1},
        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
    };

    World world = {};
    world.tileCountX = TILEMAP_WIDTH;
    world.tileCountY = TILEMAP_HEIGHT;
    world.tileMapCountX = WORLD_WIDTH;
    world.tileMapCountY = WORLD_HEIGHT;
    world.upperLeftX = 0;
    world.upperLeftY = 0;
    world.tileSideInMeters = 1.4f;
    world.tileSideInPixels = 60;

    TileMap maps[WORLD_HEIGHT][WORLD_WIDTH];
    maps[0][0].tiles = (uint32*)tiles00;

    maps[1][0] = maps[0][0];
    maps[1][0].tiles = (uint32*)tiles01;

    maps[0][1] = maps[0][0];
    maps[0][1].tiles = (uint32*)tiles10;

    maps[1][1] = maps[0][0];
    maps[1][1].tiles = (uint32*)tiles11;

    world.tileMaps = (TileMap*)maps;

    // player
    real32 playerR = 1.0f;
    real32 playerG = 1.0f;
    real32 playerB = 0.5f;
    real32 playerWidth = 50.0f;
    real32 playerHeight = 70.0f;

    for (int controllerIndex = 0; controllerIndex < ArrayCount(input->controllers); controllerIndex++) {
        GameControllerInput *controller = &input->controllers[controllerIndex];

        real32 dPlayerX = 0.0f;
        real32 dPlayerY = 0.0f;

        if (controller->stickUp.endedDown) {
            dPlayerY = -1.0f;
        }
        if (controller->stickDown.endedDown) {
            dPlayerY = 1.0f;
        }
        if (controller->stickLeft.endedDown) {
            dPlayerX = -1.0f;
        }
        if (controller->stickRight.endedDown) {
            dPlayerX = 1.0f;
        }

        dPlayerX *= 96.0f;
        dPlayerY *= 96.0f;

        CanonicalPosition newPlayerPos = state->playerPos;
        newPlayerPos.x += dPlayerX * input->dtForFrame;
        newPlayerPos.y += dPlayerY * input->dtForFrame;
        newPlayerPos = recanonicalisePosition(&world, newPlayerPos);

        CanonicalPosition playerBottomLeft = newPlayerPos;
        playerBottomLeft.x -= 0.5f * playerWidth;
        playerBottomLeft = recanonicalisePosition(&world, playerBottomLeft);

        CanonicalPosition playerTopLeft = newPlayerPos;
        playerTopLeft.x -= 0.5f * playerWidth;
        playerTopLeft.y -= playerWidth;
        playerTopLeft = recanonicalisePosition(&world, playerTopLeft);

        CanonicalPosition playerBottomRight = newPlayerPos;
        playerBottomRight.x += 0.5f * playerWidth;
        playerBottomRight = recanonicalisePosition(&world, playerBottomRight);

        CanonicalPosition playerTopRight = newPlayerPos;
        playerTopRight.x += 0.5f * playerWidth;
        playerTopRight.y -= playerWidth;
        playerTopRight = recanonicalisePosition(&world, playerTopRight);

        if (
          isWorldPointEmpty(&world, playerTopLeft) &&
          isWorldPointEmpty(&world, playerTopRight) &&
          isWorldPointEmpty(&world, playerBottomLeft) &&
          isWorldPointEmpty(&world, playerBottomRight)
        ) {
            state->playerPos = newPlayerPos;
        }
    }

    // clearscreen
    drawRectangle(videoBuf, 0.0f, 0.0f, (real32)videoBuf->width, (real32)videoBuf->height, 1.0f, 0.0f, 1.0f);

    TileMap *currMap = getTileMap(&world, state->playerPos.tileMapX, state->playerPos.tileMapY);
    for (int row = 0; row < world.tileCountY; row++) {
        for (int col = 0; col < world.tileCountX; col++) {
            real32 fill = getTileValueUnchecked(&world, currMap, col, row) == 1 ? 1.0f : 0.5f;
            real32 minX = world.upperLeftX + ((real32)col)*world.tileSideInPixels;
            real32 minY = world.upperLeftY + ((real32)row)*world.tileSideInPixels;
            real32 maxX = minX + world.tileSideInPixels;
            real32 maxY = minY + world.tileSideInPixels;
            drawRectangle(videoBuf, minX, minY, maxX, maxY, fill, fill, fill);
        }
    }

    real32 playerX = world.upperLeftX + state->playerPos.x + state->playerPos.tileX * world.tileSideInPixels;
    real32 playerY = world.upperLeftY + state->playerPos.y + state->playerPos.tileY * world.tileSideInPixels;
    real32 playerLeft = playerX - 0.5f * playerWidth;
    real32 playerTop = playerY - playerHeight;
    drawRectangle(videoBuf, playerLeft, playerTop, playerX + playerWidth*0.5f, playerY, playerR, playerG, playerB);
}

extern "C" GAME_GET_SOUND_SAMPLES(gameGetSoundSamples) {
    outputSound(soundBuf, (GameState*)memory->permanentStorage);
}
